@@ -10,9 +10,13 @@ import warnings
 
 import numpy as np
 from scipy.ndimage import median_filter, uniform_filter1d
+from scipy.signal import savgol_coeffs
 
-from ._algorithm_setup import _get_vander, _setup_smooth
-from .utils import ParameterWarning, gaussian, gaussian_kernel, padded_convolve
+from ._algorithm_setup import _get_vander, _setup_smooth, _yx_arrays
+from .utils import (
+    _check_scalar, _get_edges, ParameterWarning, gaussian, gaussian_kernel, optimize_window,
+    padded_convolve, pad_edges, relative_difference
+)
 
 
 def noise_median(data, half_window, smooth_half_window=None, sigma=None, **pad_kwargs):
@@ -157,13 +161,7 @@ def snip(data, max_half_window, decreasing=False, smooth_half_window=None,
     if filter_order not in {2, 4, 6, 8}:
         raise ValueError('filter_order must be 2, 4, 6, or 8')
 
-    if isinstance(max_half_window, int):
-        half_windows = [max_half_window, max_half_window]
-    elif len(max_half_window) == 1:
-        half_windows = [max_half_window[0], max_half_window[0]]
-    else:
-        half_windows = [max_half_window[0], max_half_window[1]]
-
+    half_windows = _check_scalar(max_half_window, 2, True, dtype=int)[0]
     num_y = len(data)
     for i, half_window in enumerate(half_windows):
         if half_window > (num_y - 1) // 2:
@@ -173,7 +171,7 @@ def snip(data, max_half_window, decreasing=False, smooth_half_window=None,
             )
             half_windows[i] = (num_y - 1) // 2
 
-    max_of_half_windows = max(half_windows)
+    max_of_half_windows = np.max(half_windows)
     if decreasing:
         range_args = (max_of_half_windows, 0, -1)
     else:
@@ -400,14 +398,14 @@ def swima(data, min_half_window=3, max_half_window=None, smooth_half_window=None
     if max_half_window is None:
         max_half_window = (len(data) - 1) // 2
     y = _setup_smooth(data, max_half_window, **pad_kwargs)
-
+    len_y = len(y)  # includes the padding of max_half_window at each side
     data_slice = slice(max_half_window, -max_half_window)
     if smooth_half_window is None:
-        smooth_half_window = max(1, y[data_slice].shape[0] // 50)
+        smooth_half_window = max(1, (len_y - 2 * max_half_window) // 50)
     if smooth_half_window > 0:
         y = uniform_filter1d(y, 2 * smooth_half_window + 1)
 
-    vander, pseudo_inverse = _get_vander(np.linspace(-1, 1, y[data_slice].shape[0] - 1), 3)
+    vander, pseudo_inverse = _get_vander(np.linspace(-1, 1, len_y - 2 * max_half_window - 1), 3)
     baseline, converged, half_window = _swima_loop(
         y, vander, pseudo_inverse, data_slice, max_half_window, min_half_window
     )
@@ -416,7 +414,7 @@ def swima(data, min_half_window=3, max_half_window=None, smooth_half_window=None
     if not converged:
         residual = y - baseline
         gaussian_bkg = gaussian(
-            np.arange(y.shape[0]), np.max(residual), y.shape[0] / 2, y.shape[0] / 6
+            np.arange(len_y), np.max(residual), len_y / 2, len_y / 6
         )
         baseline_2, converged, half_window = _swima_loop(
             residual + gaussian_bkg, vander, pseudo_inverse, data_slice, max_half_window, 3
@@ -426,3 +424,242 @@ def swima(data, min_half_window=3, max_half_window=None, smooth_half_window=None
         half_windows.append(half_window)
 
     return baseline[data_slice], {'half_window': half_windows, 'converged': converges}
+
+
+def ipsa(data, half_window=None, max_iter=500, tol=None, roi=None,
+         original_criteria=False, **pad_kwargs):
+    """
+    Iterative Polynomial Smoothing Algorithm (IPSA).
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    half_window : int
+        The half-window to use for the smoothing each iteration. Should be
+        approximately equal to the full-width-at-half-maximum of the peaks or features
+        in the data. Default is None, which will use 4 times the output of
+        :func:`.optimize_window`, which is not always a good value, but at least scales
+        with the number of data points and gives a starting point for tuning the parameter.
+    max_iter : int, optional
+        The maximum number of iterations. Default is 500.
+    tol : float, optional
+        The exit criteria. Default is None, which uses 1e-3 if `original_criteria` is
+        False, and ``1 / (max(data) - min(data))`` if `original_criteria` is True.
+    roi : slice or array-like, shape(N,)
+        The region of interest, such that ``np.asarray(data)[roi]`` gives the values
+        for calculating the tolerance if `original_criteria` is True. Not used if
+        `original_criteria` is True. Default is None, which uses all values in `data`.
+    original_criteria : bool, optional
+        Whether to use the original exit criteria from the reference, which is difficult
+        to use since it requires knowledge of how high the peaks should be after baseline
+        correction. If False (default), then compares ``norm(old, new) / norm(old)``, where
+        `old` is the previous iteration's baseline, and `new` is the current iteration's
+        baseline.
+    **pad_kwargs
+        Additional keyword arguments to pass to :func:`.pad_edges` for padding
+        the edges of the data to prevent edge effects from convolution.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    params : dict
+        A dictionary with the following items:
+
+        * 'tol_history': numpy.ndarray
+            An array containing the calculated tolerance values for
+            each iteration. The length of the array is the number of iterations
+            completed. If the last value in the array is greater than the input
+            `tol` value, then the function did not converge.
+
+    References
+    ----------
+    Wang, T., et al. Background Subtraction of Raman Spectra Based on Iterative
+    Polynomial Smoothing. Applied Spectroscopy. 71(6) (2017) 1169-1179.
+
+    """
+    # TODO should just move optimize window into _setup_smooth since all smooth functions
+    # could use it; maybe just add a multiplier constant; that way, snip and noise_median
+    # no longer require a half window parameter to at least get a guess
+    if half_window is None:
+        half_window = 4 * optimize_window(data)
+    window_size = 2 * half_window + 1
+    y = _setup_smooth(data, window_size, **pad_kwargs)
+    y0 = y
+    data_slice = slice(window_size, -window_size)
+    if original_criteria and not (roi is None or isinstance(roi, slice)):
+        roi = np.asarray(roi)
+
+    if tol is None:
+        if original_criteria:
+            # guess what the desired height should be; not a great guess, but it's
+            # something
+            tol = 1 / np.ptp(y[data_slice][roi])
+        else:
+            tol = 1e-3
+
+    savgol_coef = savgol_coeffs(window_size, 2)
+    tol_history = np.empty(max_iter + 1)
+    old_baseline = y[data_slice]
+    for i in range(max_iter + 1):
+        baseline = padded_convolve(y, savgol_coef, 'edge')
+        if original_criteria:
+            residual = (y0 - baseline)[data_slice][roi]
+            calc_tol = abs(residual.min() / residual.max())
+        else:
+            calc_tol = relative_difference(old_baseline, baseline[data_slice])
+        tol_history[i] = calc_tol
+        if calc_tol < tol:
+            break
+        y = np.minimum(y0, baseline)
+        old_baseline = baseline[data_slice]
+
+    return baseline[data_slice], {'tol_history': tol_history[:i + 1]}
+
+
+def ria(data, x_data=None, half_window=None, max_iter=500, tol=1e-2, side='both',
+        width_scale=0.1, height_scale=1., sigma_scale=1. / 12., **pad_kwargs):
+    """
+    Range Independent Algorithm (RIA).
+
+    Adds additional data to the left and/or right of the input data, and then
+    iteratively smooths until the area of the additional data is removed.
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    x_data : array-like, shape (N,), optional
+        The x-values of the measured data. Default is None, which will create an
+        array from -1 to 1 with N points.
+    half_window : int, optional
+        The half-window to use for the smoothing each iteration. Should be
+        approximately equal to the full-width-at-half-maximum of the peaks or features
+        in the data. Default is None, which will use the output of :func:`.optimize_window`,
+        which is not always a good value, but at least scales with the number of data points
+        and gives a starting point for tuning the parameter.
+    max_iter : int, optional
+        The maximum number of iterations. Default is 500.
+    tol : float, optional
+        The exit criteria. Default is 1e-2.
+    side : {'both', 'left', 'right'}, optional
+        The side of the measured data to extend. Default is 'both'.
+    width_scale : float, optional
+        The number of data points added to each side is `width_scale` * N. Default
+        is 0.1.
+    height_scale : float, optional
+        The height of the added Gaussian peak(s) is calculated as
+        `height_scale` * max(`data`). Default is 1.
+    sigma_scale : float, optional
+        The sigma value for the added Gaussian peak(s) is calculated as
+        `sigma_scale` * `width_scale` * N. Default is 1/12, which will make
+        the Gaussian span +- 6 sigma, making its total width about half of the
+        added length.
+    **pad_kwargs
+        Additional keyword arguments to pass to :func:`.pad_edges` for padding
+        the edges of the data when adding the extended left and/or right sections.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    params : dict
+        A dictionary with the following items:
+
+        * 'tol_history': numpy.ndarray
+            An array containing the calculated tolerance values for
+            each iteration. The length of the array is the number of iterations
+            completed. If the last value in the array is greater than the input
+            `tol` value, then the function did not converge (if the array length
+            is equal to `max_iter`) or the areas of the smoothed extended regions
+            exceeded their initial areas (if the array length is < `max_iter`).
+
+    Raises
+    ------
+    ValueError
+        Raised if `side` is not 'left', 'right', or 'both'.
+
+    References
+    ----------
+    Krishna, H., et al. Range-independent background subtraction algorithm for
+    recovery of Raman spectra of biological tissue. J Raman Spectroscopy. 43(12)
+    (2012) 1884-1894.
+
+    """
+    side = side.lower()
+    if side not in ('left', 'right', 'both'):
+        raise ValueError('side must be "left", "right", or "both"')
+
+    y, x = _yx_arrays(data, x_data)
+    if half_window is None:
+        half_window = optimize_window(y)
+    sort_x = x_data is not None
+    if sort_x:
+        sort_order = np.argsort(x, kind='mergesort')
+        x = x[sort_order]
+        y = y[sort_order]
+    max_x = x.max()
+    min_x = x.min()
+    x_range = max_x - min_x
+
+    added_window = int(x.shape[0] * width_scale)
+    lower_bound = 0
+    upper_bound = 0
+    known_area = 0.
+
+    # TODO should make this a function that could be used by optimizers.optimize_extended_range too
+    added_left, added_right = _get_edges(y, added_window, **pad_kwargs)
+    added_gaussian = gaussian(
+        np.linspace(-added_window / 2, added_window / 2, added_window),
+        height_scale * abs(y.max()), 0, added_window * sigma_scale
+    )
+    if side in ('left', 'both'):
+        added_x_left = np.linspace(
+            min_x - x_range * (width_scale / 2), min_x, added_window + 1
+        )[:-1]
+        added_y_left = added_gaussian + added_left
+        lower_bound = added_window
+        known_area += np.trapz(added_gaussian, added_x_left)
+    else:
+        added_x_left = []
+        added_y_left = []
+
+    if side in ('right', 'both'):
+        added_x_right = np.linspace(
+            max_x, max_x + x_range * (width_scale / 2), added_window + 1
+        )[1:]
+        added_y_right = added_gaussian + added_right
+        upper_bound = added_window
+        known_area += np.trapz(added_gaussian, added_x_right)
+    else:
+        added_x_right = []
+        added_y_right = []
+
+    fit_x_data = np.concatenate((added_x_left, x, added_x_right))
+    fit_data = np.concatenate((added_y_left, y, added_y_right))
+
+    upper_max = fit_data.shape[0] - upper_bound
+    tol_history = np.empty(max_iter)
+    window_size = 2 * half_window + 1
+    smoother_array = pad_edges(fit_data, window_size, extrapolate_window=2)
+    data_slice = slice(window_size, -window_size)
+    for i in range(max_iter):
+        # only smooth fit_data so that the outer section remains unchanged by
+        # smoothing and edge effects are ignored
+        smoother_array[data_slice] = uniform_filter1d(smoother_array, window_size)[data_slice]
+        residual = fit_data - smoother_array[data_slice]
+        calc_area = np.trapz(residual[:lower_bound], fit_x_data[:lower_bound])
+        if upper_bound:
+            calc_area += np.trapz(residual[-upper_bound:], fit_x_data[-upper_bound:])
+        calc_difference = relative_difference(known_area, calc_area)
+        tol_history[i] = calc_difference
+        if calc_difference < tol or calc_area > known_area:
+            break
+        smoother_array[data_slice] = np.minimum(fit_data, smoother_array[data_slice])
+
+    baseline = smoother_array[data_slice][lower_bound:upper_max]
+    if sort_x:
+        baseline = baseline[np.argsort(sort_order, kind='mergesort')]
+
+    return baseline, {'tol_history': tol_history[:i + 1]}
