@@ -40,3 +40,13 @@ def snip(*args, **kwargs):  # noqa
 @_wrap_and_warn
 def swima(*args, **kwargs):  # noqa
     return smooth.swima(*args, **kwargs)
+
+
+@_wrap_and_warn
+def ipsa(*args, **kwargs):  # noqa
+    return smooth.ipsa(*args, **kwargs)
+
+
+@_wrap_and_warn
+def ria(*args, **kwargs):  # noqa
+    return smooth.ria(*args, **kwargs)
